package org.processmining.plugins.DataConformance.DataAlignment.DCNet;

import java.text.ParseException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;
import java.util.Map.Entry;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import net.sf.javailp.Linear;
import net.sf.javailp.OptType;
import net.sf.javailp.Problem;
import net.sf.javailp.Result;
import net.sf.javailp.Solver;
import net.sf.javailp.SolverFactory;
import net.sf.javailp.SolverFactoryLpSolve;
import net.sf.javailp.VarType;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.models.DCNet.ActivityBinding;
import org.processmining.models.DCNet.DCNet;
import org.processmining.models.DCNet.DCNetNode;
import org.processmining.plugins.DataConformance.Utility;
import org.processmining.plugins.DataConformance.DataAlignment.DataAlignmentState;
import org.processmining.plugins.DataConformance.DataAlignment.GenericTrace;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.framework.ReplayState;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;
import org.processmining.plugins.cnet.replayresult.CNetRepResult;

public class MultiThreadedDataAlignment implements Runnable
{
	private Condition current;
	private Map<String, String> variableMapping;
	private XTrace trace;
	private VariableMatchCosts variableCost;
	private UIPluginContext context;
	private Collection<ReplayState> list;
	private HashMap<DCNetNode,XEventClass> activityMapping;
	private XEventClasses eventClasses;
	private DCNet net;
	
	public MultiThreadedDataAlignment(DCNet net, Condition current, HashMap<DCNetNode, XEventClass> activityMapping, Map<String, String> variableMapping, XTrace trace, VariableMatchCosts variableCost, UIPluginContext context, XEventClasses eventClasses, Collection<ReplayState> list)
	{
		this.current=current;
		this.variableMapping=variableMapping;
		this.trace=trace;
		this.variableCost=variableCost;
		this.activityMapping=activityMapping;
		this.context=context;
		this.list=list;
		this.eventClasses=eventClasses;
		this.net=net;
	}
	
	
	public void run()
	{
		System.out.println(XConceptExtension.instance().extractName(trace));
		Linear objective=new Linear();
		Iterator<XEvent> traceIterator=trace.iterator();

		float additionalCost=0; // the cost of all variables that were not written whereas they were supposed to. 
		
		//Map of the process var Name onto the variable name
		Iterator<HashMap<String,String>> varWrittenIterator=current.varWrittens.iterator();
		Iterator<ActivityBinding> bindingIterator=current.activitySet.iterator();
		Problem problem=current.getIlpProblem(false);
		boolean logMoves[]=current.logMove;	
		for(int i=0;i<logMoves.length;i++)
		{
			ActivityBinding binding=bindingIterator.next();
			HashMap<String,String> varWritten=null;
			if (binding.node!=null)
				varWritten=varWrittenIterator.next();
			if(logMoves[i]) 
			{
				XEvent event=traceIterator.next();
				if (binding.node!=null)
					additionalCost+=addLogMove(problem,event,objective,varWritten);
			}
		}
		problem.setObjective(objective,OptType.MIN);
		DataAlignmentState state=createAlignmentState(problem,additionalCost);
		if (state!=null)
		{
			state.setControlFlowFitness(current.getControlFlowFitness());
			list.add(state);
		}
		else
			context.log("The Casual Net contains errors in the data flow. Impossible to find a data alignment for trace "+
					XConceptExtension.instance().extractName(trace));
			
		synchronized(context) {
			context.getProgress().inc();
		}
	}
	
	private ExecutionStep getExecutionStep(Result result, HashMap<String,String> variables, String activityName)
	{
		ExecutionStep step=new ExecutionStep(activityName);
		for(Entry<String, String> entry: variables.entrySet())
		{
			Number value=result.get(entry.getValue());
			Class<?> classType=net.getSetValues(net.getSetForVariable(entry.getKey())).iterator().next().getClass();
			step.put(entry.getKey(), org.processmining.models.guards.NumericValueConversion.fromNumericValue(value.doubleValue(), classType));
		}
		return step;
	}
	
	private ExecutionStep getExecutionStep(XEvent event,String stepName)
	{
		if (stepName==null)
			stepName=XConceptExtension.instance().extractName(event);
		ExecutionStep step=new ExecutionStep(stepName);
		for(Entry<String, String> entr : variableMapping.entrySet()) {
			if (stepName.equals(entr.getValue())) {
				stepName=entr.getKey();
				break;
			}
		}		
		for(Entry<String, XAttribute> attribute : event.getAttributes().entrySet())
		{ 
			Object value=Utility.getValue(attribute.getValue());
			if (value!=null)
			{
				for(Entry<String, String> entr : variableMapping.entrySet()) {
					if (attribute.getKey().equals(entr.getValue())) {
						step.put(entr.getKey(),value);
						break;
					}
				}
			}
		}
		return step;		
	}
	
	private DataAlignmentState createAlignmentState(Problem problem, float additionalCost)
	{
		SolverFactory factory = new SolverFactoryLpSolve(); // use lp_solve
		factory.setParameter(Solver.VERBOSE, 0);
		factory.setParameter(Solver.TIMEOUT, 100); // set timeout to 100 seconds
		Solver solver = factory.get(); 
		Result result = solver.solve(problem);
		System.out.println(problem);
		System.out.println(result);
		if (result==null)
			return null;
		Number objectiveValue=result.getObjective();
			
		boolean logMoves[]=current.logMove;
		Iterator<ActivityBinding> activityBindingIterator=current.activitySet.iterator();
		String traceName=XConceptExtension.instance().extractName(trace);
		
		GenericTrace logTrace=new GenericTrace(logMoves.length,	traceName);
		GenericTrace processTrace=new GenericTrace(logMoves.length,"");
		
		
		Iterator<XEvent> traceIterator=trace.iterator();
		//Map of the process var name onto the variable name
		Iterator<HashMap<String,String>> varWrittenIterator=current.varWrittens.iterator();
		int stepIndex=0;
		while(activityBindingIterator.hasNext())
		{
			ActivityBinding binding=activityBindingIterator.next();
			String activityName=null;
			if (binding.node!=null) {
				ExecutionStep step=getExecutionStep(result, varWrittenIterator.next(), binding.node.getLabel());
				processTrace.add(step);
				activityName=step.getActivity();
			}
			else
				processTrace.add(ExecutionStep.bottomStep);
			if (logMoves[stepIndex++]) 
				logTrace.add(getExecutionStep(traceIterator.next(),activityName));
			else
				logTrace.add(ExecutionStep.bottomStep);		
		}
				
		return new DataAlignmentState(logTrace, processTrace,objectiveValue.floatValue());		
	}
	
	private String getProcessActivityName(XEvent event)
	{
		for(Entry<DCNetNode, XEventClass> entry : activityMapping.entrySet())
		{
			if (eventClasses.getClassOf(event).equals(entry.getValue()))
				return entry.getKey().getLabel();
		}
		return null;
	}

	private float addLogMove(Problem problem,XEvent event,Linear objective, HashMap<String,String> varWritten)
	{
		float additionalCost=0;
		for(Entry<String,String> entry : varWritten.entrySet())
		{
			String processVarName = entry.getKey();
			String logVarName=variableMapping.get(processVarName);
			XAttribute attr=event.getAttributes().get(logVarName);
			String ilpVarName=entry.getValue();
			if (attr!=null)
			{
				String boolVar=ilpVarName+"_D";
				problem.setVarType(boolVar, Boolean.class);
				Object value=Utility.getValue(attr);
				double attrVal=org.processmining.models.guards.NumericValueConversion.toNumericValue(value);
				double upperbound;
				if (problem.getVarType(ilpVarName)==VarType.BOOL)
					upperbound=1;
				else
					upperbound=problem.getVarUpperBound(ilpVarName).doubleValue();
				Linear constraint=new Linear();
				constraint.add(1, ilpVarName);
				constraint.add(-1.0*upperbound,boolVar);
				problem.add(constraint, "<=", attrVal);
				constraint=new Linear();
				constraint.add(-1, ilpVarName);
				constraint.add(-1.0*upperbound,boolVar);
				problem.add(constraint, "<=", -1.0*attrVal);
				objective.add(variableCost.costFaultyValue(getProcessActivityName(event), processVarName), boolVar);
			}
			else
				additionalCost+=variableCost.notWritingCost(getProcessActivityName(event), processVarName);
		}
		return additionalCost;
	}
	
	public static Collection<ReplayState> perform(DCNet net, XLog log,CNetRepResult input, HashMap<DCNetNode, XEventClass> activityMapping, XEventClasses eventClasses, Map<String,String> variableMapping, VariableMatchCosts variableCost, UIPluginContext context) throws InterruptedException, ParseException
	{		
		int maxConcurrentThreads=Runtime.getRuntime().availableProcessors();
		Collection<ReplayState> list=new Vector<ReplayState>();
		System.loadLibrary("lpsolve55");
		System.loadLibrary("lpsolve55j");
		
		Map<Integer,Condition> conditions=Condition.getConditions(net, input);
		
		ThreadPoolExecutor pool = new ThreadPoolExecutor(maxConcurrentThreads, maxConcurrentThreads, 60, TimeUnit.SECONDS,new LinkedBlockingQueue<Runnable>());
		pool.allowCoreThreadTimeOut(true);
		int traceIndex=0;
		for (XTrace trace : log)
		{
			Condition current=conditions.get(traceIndex++);
			MultiThreadedDataAlignment aligner=
				new MultiThreadedDataAlignment(net, current, activityMapping, variableMapping, trace,variableCost, context, eventClasses, list);
			pool.execute(aligner);
		}
		
		pool.shutdown();
		pool.awaitTermination(10, TimeUnit.DAYS);
			
		return list;
		
	}	

}
