package org.processmining.plugins.DataConformance;

import java.util.Collection;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.models.DCNet.DCNet;
import org.processmining.models.DCNet.DCNetNode;
import org.processmining.plugins.DataConformance.DCnet.ResultReplayDCNet;
import org.processmining.plugins.DataConformance.GUI.ActivityMatchCostPanel;
import org.processmining.plugins.DataConformance.GUI.MappingPanel;
import org.processmining.plugins.DataConformance.GUI.VariableMatchCostPanel;
import org.processmining.plugins.DataConformance.framework.ActivityMatchCosts;
import org.processmining.plugins.DataConformance.framework.BelongsTo;
import org.processmining.plugins.DataConformance.framework.ReplayState;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;

public class DataConformanceChecking 
{
	
	
/*	@SuppressWarnings("unchecked")
	@Plugin(name = "Conformance Checking of Data-aware Casual Net (using A* algorithm)",
			parameterLabels = {"Event log", "Data-aware Casual Net"},
			returnLabels = { "Replay Result" },
			returnTypes = { ResultReplayDCNet.class },
			userAccessible = true,
			help = "Data Conformance")
			@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "M. de Leoni", email = "m.d.leoni@tue.nl")		
	*/
	public static ResultReplayDCNet dataConformance(UIPluginContext context, XLog log, DCNet model) throws InterruptedException 
	{
		XLogInfo summary = XLogInfoFactory.createLogInfo(log);
		XEventClasses eventClasses=summary.getEventClasses();
		TreeSet<DCNetNode> DCnodes=new TreeSet<DCNetNode>((Set)model.getNodes());
		BelongsTo.getInstance().resetForDCNet(model);

		//Activity Mapping
		MappingPanel<DCNetNode, XEventClass> mapActivityPanel=new MappingPanel<DCNetNode, XEventClass>(DCnodes,eventClasses.getClasses());
		InteractionResult result = context.showConfiguration("Setup Activity Mapping", mapActivityPanel);
		if (result == InteractionResult.CANCEL) {
			context.getFutureResult(0).cancel(true);
			return null;
		}
		Map<DCNetNode,XEventClass> activityMapping=mapActivityPanel.getMapping(false);
	
		//Variable Mapping
		Collection<String> logAttributes=summary.getEventAttributeInfo().getAttributeKeys();
		TreeSet<String> processAttributes=new TreeSet<String>(model.getVariables());
	
		MappingPanel<String, String> mapVariablePanel=
			new MappingPanel<String, String>(processAttributes,logAttributes);
		result = context.showConfiguration("Setup Variable Mapping", mapVariablePanel);
		if (result == InteractionResult.CANCEL) {
			context.getFutureResult(0).cancel(true);
			return null;
		}
		Map<String, String> variableMapping=mapVariablePanel.getMapping(false);		
	
		//Activity Costs
		ActivityMatchCostPanel<XEventClass> activityPanel=new ActivityMatchCostPanel<XEventClass>(DCnodes,activityMapping);		
		result = context.showConfiguration("Activity Mapping Cost", activityPanel);
		if (result == InteractionResult.CANCEL) {
			context.getFutureResult(0).cancel(true);
			return null;
		}	
		ActivityMatchCosts<XEventClass> activityCost=activityPanel.getCosts();
		
		//Variable Costs
		VariableMatchCostPanel<XEventClass> variablePanel=new VariableMatchCostPanel<XEventClass>(activityMapping,mapVariablePanel.getMapping(true));
		result = context.showConfiguration("Variable Mapping Cost", variablePanel);
		if (result == InteractionResult.CANCEL) {
			context.getFutureResult(0).cancel(true);
			return null;
		}

		VariableMatchCosts variableCost=variablePanel.getCosts();
		Map<String,Float> maxDistance=
			org.processmining.plugins.DataConformance.Utility.getMaxDistanceInValue(model, log, variableMapping);
		
		Progress progBar = context.getProgress();
		progBar.setMinimum(0);
		progBar.setMaximum(log.size());
		progBar.setValue(0);
		
		long beforeTime=System.currentTimeMillis();
		
		Collection<ReplayState> list=
			org.processmining.plugins.DataConformance.MultiThreadedConformanceChecking.perform(model, activityMapping, variableMapping, eventClasses, log, activityCost, variableCost, maxDistance, context, true);
				
		long afterTime=System.currentTimeMillis();
		
		context.log("Computation time: "+(afterTime-beforeTime)+" secs.");
		ResultReplayDCNet returnedObject=new ResultReplayDCNet(list,activityCost,variableCost,model,log,variableMapping, eventClasses.getClassifier());
		context.log("Visited Nodes: "+returnedObject.getMeanVisitedNode());
		context.log("Size of the built tree: "+returnedObject.getMeanBuiltTreeSize());
		
		return returnedObject;
	}
	
}