package org.processmining.plugins.DataConformance.GUI;

import java.awt.Component;
import java.util.Arrays;
import java.util.Collection;
import java.util.Set;
import java.util.TreeSet;

import javax.swing.AbstractCellEditor;
import javax.swing.JComboBox;
import javax.swing.JTable;
import javax.swing.table.TableCellEditor;

import org.processmining.plugins.DataConformance.framework.ReplayableActivity;

class MyTableCellEditor extends AbstractCellEditor implements TableCellEditor 
{
	private Object[] values;
	private JComboBox allValuesCombo;
	private JComboBox current;
	
	public MyTableCellEditor(Object[] values)
	{
		this.values=values;
		this.allValuesCombo=new JComboBox(values);
		allValuesCombo.addItem("");
	}
	
    // This method is called when a cell value is edited by the user.
    public Component getTableCellEditorComponent(JTable table, Object value,
            boolean isSelected, int rowIndex, int vColIndex) 
    {
    	int otherColIndex=(vColIndex==0 ? 1 : 0);
    	Object valueAt=table.getModel().getValueAt(rowIndex, otherColIndex);
    	if (valueAt.equals(""))
    	{
    		current=allValuesCombo;
    		allValuesCombo.setSelectedItem(value);
    		return allValuesCombo;
    	}
    	else
    	{
    		current=new JComboBox();
    		current.addItem("");
    		current.setSelectedIndex(0);
    		return current;
    	}
    }

    public Object getCellEditorValue() {
        return current.getSelectedItem();
    }
}

public class ActivityMatchingCostTable<E> extends JTable
{
	ActivityMatchingTableModel model;
	

	public ActivityMatchingCostTable(Set<? extends ReplayableActivity> activities,Collection<E> eventClasses) 
	{
		super(new ActivityMatchingTableModel());
		this.model=(ActivityMatchingTableModel) this.getModel();
		TreeSet<E> app=new TreeSet<E>();
		app.addAll(eventClasses);
		Object[] activitiesArray=activities.toArray();
		Arrays.sort(activitiesArray);
		this.getColumnModel().getColumn(1).setCellEditor(new MyTableCellEditor(activitiesArray));
		this.getColumnModel().getColumn(0).setCellEditor(new MyTableCellEditor(app.toArray()));	
	}
	
	public void addRow()
	{
		if (editingRow==-1)
		{	
			int selectedRow=this.getSelectedRow();
			if (selectedRow<0) selectedRow=0;
			model.addRow(selectedRow);
			model.fireTableDataChanged();
		}
	}
	
	public void removeRow()
	{
		if (editingRow==-1)
		{
			int[] indexes=this.getSelectedRows();
			model.removeRow(indexes);
			model.fireTableDataChanged();
		}
	}
	
	public boolean rowBeingEdited()
	{
		return (editingRow>-1);
	}
}
