package org.processmining.plugins.DataConformance.framework;

import java.io.IOException;
import java.util.AbstractMap.SimpleEntry;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.PriorityQueue;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;

public class DataReplayer extends Thread
{
	private Replayable model;

	private XTrace trace;
	private Collection<ReplayState> list;
	private HashMap<SimpleEntry<ExecutionTrace, ExecutionTrace>,Float> visitedNode=new HashMap<SimpleEntry<ExecutionTrace, ExecutionTrace>,Float>();
	private UIPluginContext bar;
	protected ReplayingHelper helper;
	public int visitedNodes=0;
	public long treeSize=0;

	private boolean prune;		

	public DataReplayer(Replayable model, Map<ReplayableActivity, XEventClass> activityMapping,
			Map<String, String> variableMapping, XEventClasses eventClasses,XTrace trace,
			ActivityMatchCosts activityCost, VariableMatchCosts variableCost, 
			Map<String, Float> maxDistance, Collection<ReplayState> list, UIPluginContext context, boolean prune) 
	{
		this.model=model;
		this.trace=trace;
		this.list=list;
		this.bar=context;
		helper=new ReplayingHelper(model,trace,activityMapping,variableMapping,eventClasses, 
				maxDistance, activityCost, variableCost);
		this.prune=prune;
	}
	
	public void run()
	{
		try
		{
			ReplayState replayingState = perform();
			if (replayingState==null)
				bar.log("Impossible to replay the model: is the model sound?");
			else
				list.add(replayingState);
		}
		catch(Exception err)
		{
			synchronized(bar) {
				bar.log("Error while replaying trace "+trace.getAttributes().get("concept:name")+": "+err.getMessage());
			}
			err.printStackTrace();
		}
		synchronized(bar) {
			bar.log("Trace "+trace.getAttributes().get("concept:name")+" replayed");
			bar.getProgress().inc();
		}
		synchronized(model) {			
			model.notify();
		}
	}

	
	public ReplayState perform() throws IOException 
	{
		ExecutionTrace sigmaT=helper.buildLogSequence();
		synchronized(bar) {
			bar.log("Starting replay trace "+trace.getAttributes().get("concept:name"));
		}		
		PriorityQueue<ReplayState> openList=new PriorityQueue<ReplayState>();	
		ReplayState mu0=new ReplayState();
		mu0.processTracePrefix=model.buildEmptyPrefix();
		mu0.logTracePrefix=model.createLogExecutionTrace(sigmaT);
		
		ReplayState mu=mu0;

		while(mu!=null && (!mu.processTracePrefix.isComplete() || !mu.logTracePrefix.isComplete()))
		{
			Collection<ReplayState> successors=helper.createSuccessors(mu);
			for(ReplayState mu1 : successors)
			{
				mu1.h_cost=helper.computeHeuristicValue(mu1,sigmaT);
				float mu_g_cost=mu.f_cost-mu.h_cost;
				mu1.f_cost=mu_g_cost+helper.computeDistance(mu,mu1)+mu1.h_cost;
				if (!prune || toBeVisited(mu1))
				{
					openList.add(mu1);
					treeSize++;
				}
			}
			mu=openList.poll();
			visitedNodes++;
		}
		openList.clear();
		if (mu!=null)
		{
			mu.visitedNodes=visitedNodes;
			mu.treeSize=treeSize;
		}
		return mu;
	}

	private boolean toBeVisited(ReplayState mu1) {
		SimpleEntry<ExecutionTrace, ExecutionTrace> entry=
			new SimpleEntry<ExecutionTrace, ExecutionTrace>(((LogExecutionTrace)mu1.logTracePrefix).removeBottomEvent(), mu1.processTracePrefix);
		float gValueForMu1=mu1.f_cost-mu1.h_cost;
		Float gValue=visitedNode.get(entry);
		if (gValue==null || gValueForMu1<gValue)
		{
			visitedNode.put(entry, gValueForMu1);
			return true;
		}
		else 
			return false;
	}
	

}
