package org.processmining.plugins.DataConformance.visualization.alignment;

import java.awt.Color;
import java.text.DecimalFormat;
import java.util.Iterator;
import java.util.Map;

import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Event;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Trace;
import org.processmining.xesalignmentextension.XAlignmentExtension.MoveType;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignmentMove;
import org.processmining.xesalignmentextension.XDataAlignmentExtension;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.DataMoveType;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.XDataAlignmentMove;

import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.collect.Iterators;
import com.google.common.collect.UnmodifiableIterator;

public final class AlignmentTrace implements Trace<Event> {
	
	public enum InvisibleSetting {
		HIDDEN, VISIBLE
	}
	
	public enum DeviationsSetting {
		NORMAL, HIGHLIGHTED
	}

	private static final DecimalFormat FLOAT_FORMAT = new DecimalFormat("##0.00%");

	private final XAlignment alignment;
	private final DeviationsSetting deviations;
	private final InvisibleSetting invisible;

	private boolean colorCodeActivities;
	private final Map<String, Color> activityColorMap;

	public AlignmentTrace(final XAlignment alignment, Map<String, Color> activityColorMap) {
		this(alignment, activityColorMap, InvisibleSetting.VISIBLE, DeviationsSetting.NORMAL, false);
	}
	
	public AlignmentTrace(final XAlignment alignment, Map<String, Color> activityColorMap, final InvisibleSetting invisible) {
		this(alignment, activityColorMap, invisible, DeviationsSetting.NORMAL, false);
	}

	public AlignmentTrace(final XAlignment alignment, Map<String, Color> activityColorMap, final InvisibleSetting invisible, final DeviationsSetting deviations, boolean colorCodeActivities) {
		this.alignment = alignment;
		this.activityColorMap = activityColorMap;
		this.deviations = deviations;
		this.invisible = invisible;
		this.colorCodeActivities = colorCodeActivities;
	}

	public String getName() {
		String name = alignment.getName();
		return "Trace: ".concat(name != null ? name : "unkown trace");
	}

	public Color getNameColor() {
		return null;
	}

	public String getInfo() {
		return "Fitness ".concat(FLOAT_FORMAT.format(getAlignment().getFitness()));
	}

	public Color getInfoColor() {
		return null;
	}

	public Iterator<Event> iterator() {
		UnmodifiableIterator<XAlignmentMove> filteredTrace = Iterators.filter(alignment.iterator(), new Predicate<XAlignmentMove>() {
			
			public boolean apply(XAlignmentMove step) {
				if ((getInvisible() == InvisibleSetting.HIDDEN && !step.isObservable())) {
					return false;
				}				
				return true;
			}
			
		} );
		return Iterators.transform(filteredTrace, new Function<XAlignmentMove, Event>() {

			public Event apply(XAlignmentMove move) {
				final XDataAlignmentMove dataMove = XDataAlignmentExtension.instance().extendXAlignmentMove(move);
				return new XAlignmentEvent(dataMove, activityColorMap, colorCodeActivities) {

					public String getTopLabel() {
						if (deviations == DeviationsSetting.HIGHLIGHTED && dataMove.getType() == MoveType.SYNCHRONOUS && dataMove.getDataMoveType() == DataMoveType.CORRECT) {							
							return "";	
						} else {
							return super.getTopLabel();
						}
					}

					public String getBottomLabel() {
						if (deviations == DeviationsSetting.HIGHLIGHTED && dataMove.getType() == MoveType.SYNCHRONOUS && dataMove.getDataMoveType() == DataMoveType.CORRECT) {							
							return "";	
						} else {
							return super.getBottomLabel();
						}
					}

					public String getBottomLabel2() {
						if (deviations == DeviationsSetting.HIGHLIGHTED && dataMove.getType() == MoveType.SYNCHRONOUS && dataMove.getDataMoveType() == DataMoveType.CORRECT) {							
							return "";	
						} else {
							return super.getBottomLabel2();
						}
					}

					public String getLabel() {
						String label = super.getLabel();
						if (deviations == DeviationsSetting.HIGHLIGHTED && dataMove.getType() == MoveType.SYNCHRONOUS && dataMove.getDataMoveType() == DataMoveType.CORRECT) {							
							return label != null && label.length() > 0 ? label.substring(0,1) : label;	
						} else {
							return label;
						}
					}
					
				};
			}

		});
	}

	public DeviationsSetting getDeviations() {
		return deviations;
	}

	public InvisibleSetting getInvisible() {
		return invisible;
	}

	public XAlignment getAlignment() {
		return alignment;
	}

}