package org.processmining.plugins.DataConformance.visualization.alignment;

import java.awt.Color;
import java.util.HashMap;
import java.util.Map;

import org.processmining.framework.util.ui.widgets.ColorScheme;
import org.processmining.plugins.DataConformance.visualization.DataAwareStepTypes;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignmentMove;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.DataMoveType;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.XDataAlignmentMove;

import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;
import com.google.common.collect.Multiset.Entry;
import com.google.common.collect.Multisets;

public enum ColorTheme {

	DARK("Dark Theme") {

		public Color getForeground() {
			return Color.WHITE;
		}

		public Color getBackground() {
			return Color.BLACK;
		}

	},

	BRIGHT("Bright Theme") {

		public Color getForeground() {
			return Color.BLACK;
		}

		public Color getBackground() {
			return Color.WHITE;
		}

	};

	private String description;

	private ColorTheme(String description) {
		this.description = description;
	}

	public abstract Color getForeground();

	public abstract Color getBackground();

	@Override
	public String toString() {
		return description;
	}

	public static Color getMoveColor(XDataAlignmentMove move) {
		switch (move.getType()) {
			case LOG :
				return DataAwareStepTypes.L.getColor();
			case MODEL :
				if (!move.isObservable()) {
					if (move.getDataMoveType() == DataMoveType.INCORRECT) {
						return Color.WHITE;
					}
					return DataAwareStepTypes.MINVI.getColor();
				} else {
					return DataAwareStepTypes.MREAL.getColor();
				}
			case SYNCHRONOUS :
				switch (move.getDataMoveType()) {
					case CORRECT :
						return DataAwareStepTypes.LMGOOD.getColor();
					case INCORRECT :
						return DataAwareStepTypes.LMNOGOOD.getColor();
					case MISSING :
						return DataAwareStepTypes.LMNOGOOD.getColor();
				}
		}
		throw new IllegalArgumentException("Missing Color!");
	}

	public static Map<String, Color> createColorMap(Iterable<XAlignment> alignments) {
		Map<String, Color> activityColorMap = new HashMap<>();
		Multiset<String> activityFrequency = HashMultiset.create();
		for (XAlignment a : alignments) {
			for (XAlignmentMove move : a) {				
				String activityId = move.getActivityId();
				boolean observable = move.isObservable();
				if (observable && activityId != null) {
					activityFrequency.add(activityId);
				}
			}
		}
		int i = 0;
		for (Entry<String> activityEntry : Multisets.copyHighestCountFirst(activityFrequency).entrySet()) {
			activityColorMap.put(activityEntry.getElement(), ColorScheme.COLOR_BREWER_12CLASS_PAIRED.getColor(i++));
		}
		return activityColorMap;
	}

}