package org.processmining.plugins.DataConformance.visualization.comparison;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;

import javax.swing.BoxLayout;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JPanel;

import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.widgets.ProMSplitPane;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList.TraceBuilder;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Event;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Trace;
import org.processmining.plugins.DataConformance.visualization.comparison.AlignmentComparisonPlugin.AlignmentComparisonResult;
import org.processmining.plugins.DataConformance.visualization.comparison.AlignmentComparisonPlugin.AlignmentEntry;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import com.google.common.collect.ComparisonChain;

public class AlignmentComparisonVisualizer {

	private enum ComparisonOrdering implements Comparator<AlignmentEntry> {
		DISTANCE_ASC("Distance (Ascending)") {
			@Override
			public int compare(final AlignmentEntry a, final AlignmentEntry b) {
				return ComparisonChain.start().compare(a.getDistance(), b.getDistance()).result();
			}
		},
		DISTANCE_DESC("Distance (Descending)") {
			@Override
			public int compare(final AlignmentEntry a, final AlignmentEntry b) {
				return Collections.reverseOrder(DISTANCE_ASC).compare(a, b);
			}
		},
		NAME_ASC("Name (Ascending)") {
			@Override
			public int compare(final AlignmentEntry a, final AlignmentEntry b) {
				return ComparisonChain.start().compare(a.getName(), b.getName()).result();
			}
		},
		NAME_DESC("Name (Descending)") {
			@Override
			public int compare(final AlignmentEntry a, final AlignmentEntry b) {
				return Collections.reverseOrder(NAME_ASC).compare(a, b);
			}
		},
		LENGTH_ASC("Length (Ascending)") {
			@Override
			public int compare(final AlignmentEntry a, final AlignmentEntry b) {
				return ComparisonChain.start().compare(a.getMaxLength(), b.getMaxLength()).result();
			}
		},
		LENGTH_DESC("Length (Descending)") {
			@Override
			public int compare(final AlignmentEntry a, final AlignmentEntry b) {
				return Collections.reverseOrder(LENGTH_ASC).compare(a, b);
			}
		};

		private String description;

		private ComparisonOrdering(String description) {
			this.description = description;
		}

		@Override
		public String toString() {
			return description;
		}
	}

	private ProMTraceList<AlignmentEntry> mainView;

	@Plugin(name = "Visualise Log Comparison", level = PluginLevel.NightlyBuild, //
			returnLabels = { "Log Comparison" }, returnTypes = { JComponent.class }, userAccessible = true, parameterLabels = { "Log Comparison Result" })
	@Visualizer
	@PluginVariant(requiredParameterLabels = { 0 })
	public JComponent visualise(PluginContext context, final AlignmentComparisonResult comparisonResult) throws Exception {

		context.getProgress().setCaption("Visualising ...");

		ProMSplitPane splitPane = new ProMSplitPane(ProMSplitPane.VERTICAL_SPLIT);
		splitPane.setResizeWeight(1.0d);

		JPanel topPanel = new JPanel();
		topPanel.setOpaque(true);
		topPanel.setBackground(Color.WHITE);
		topPanel.setLayout(new BoxLayout(topPanel, BoxLayout.Y_AXIS));
		mainView = createMainView(comparisonResult.getTraceEntries());
		topPanel.add(mainView);

		JPanel bottomPanel = new JPanel();
		bottomPanel.setOpaque(true);
		bottomPanel.setBackground(Color.WHITE);
		bottomPanel.setLayout(new BoxLayout(bottomPanel, BoxLayout.X_AXIS));
		bottomPanel.add(createSortComboBox());

		splitPane.setTopComponent(topPanel);
		splitPane.setBottomComponent(bottomPanel);
		return splitPane;
	}

	@SuppressWarnings("rawtypes")
	private JComboBox createSortComboBox() {
		final JComboBox groupSortBox = SlickerFactory.instance().createComboBox(ComparisonOrdering.values());
		groupSortBox.setSelectedItem(ComparisonOrdering.NAME_ASC);
		groupSortBox.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				ComparisonOrdering ordering = (ComparisonOrdering) groupSortBox.getSelectedItem();
				mainView.sort(ordering);
			}
		});
		groupSortBox.setPreferredSize(new Dimension(150, 30));
		return groupSortBox;
	}

	private ProMTraceList<AlignmentEntry> createMainView(Collection<AlignmentEntry> traceEntries) {
		final ProMTraceList<AlignmentEntry> mainView = new ProMTraceList<>(traceEntries,
				new TraceBuilder<AlignmentEntry>() {

					public Trace<? extends Event> build(AlignmentEntry element) {
						return element.getCombinedTrace();
					}
				}, ComparisonOrdering.NAME_ASC);
		mainView.setOpaque(true);
		mainView.setBackground(Color.WHITE);
		mainView.setForeground(Color.BLACK);
		mainView.setPreferredSize(new Dimension(800, Short.MAX_VALUE));

		return mainView;
	}

}
