package org.processmining.plugins.DataConformance.visualization.grouping;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceList;
import org.processmining.framework.util.ui.widgets.traceview.masterdetail.MasterView;
import org.processmining.plugins.DataConformance.visualization.alignment.XTraceResolver;
import org.processmining.plugins.DataConformance.visualization.grouping.GroupedAlignments.AlignmentGroup;

public class GroupedAlignmentMasterView<T> extends JPanel implements MasterView<AlignmentGroup, T> {

	public final static class GroupedAlignmentInput<T> {

		public final Map<String, Color> activityColorMap;
		public final GroupedAlignments<T> groupedResult;
		public final XTraceResolver traceResolver;

		public GroupedAlignmentInput(GroupedAlignments<T> groupedResult, XTraceResolver traceResolver,
				Map<String, Color> activityColorMap) {
			this.groupedResult = groupedResult;
			this.traceResolver = traceResolver;
			this.activityColorMap = activityColorMap;
		}

	}

	private static final long serialVersionUID = -6796573934109608971L;

	private GroupedAlignments<T> groupedResult;
	private Comparator<AlignmentGroup> currentOrdering = GroupedAlignments.GroupOrdering.COUNT_DESC;
	private GroupedAlignmentView groupedAlignmentsView;

	public GroupedAlignmentMasterView(PluginContext context, GroupedAlignmentInput<T> input) {
		super(new BorderLayout());
		setOpaque(false);
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		setForeground(null);
		setBackground(null);

		add(createHeading("GROUPED ALIGNMENTS VARIANTS"), BorderLayout.NORTH);

		this.groupedResult = input.groupedResult;
		groupedAlignmentsView = new GroupedAlignmentView(input.groupedResult.getAlignmentGroups(),
				currentOrdering);
		groupedAlignmentsView.showToolbar();
		add(groupedAlignmentsView, BorderLayout.CENTER);
	}

	public void clear() {
		getListView().clear();
	}

	public JComponent getMasterComponent() {
		return this;
	}

	public ProMTraceList<AlignmentGroup> getMasterList() {
		return getListView();
	}

	public Collection<T> getDetailElements(AlignmentGroup element) {
		return groupedResult.listAlignments(element);
	}

	public Comparator<AlignmentGroup> getCurrentOrdering() {
		return currentOrdering;
	}

	public void setCurrentOrdering(Comparator<AlignmentGroup> ordering) {
		currentOrdering = ordering;
		getListView().sort(currentOrdering);
	}

	public void replace(GroupedAlignments<T> newResult) {
		// Remember current selection as Set of AlignmentGroup as the "count" of an Entry may change
		@SuppressWarnings("rawtypes")
		JList masterJList = getListView().getList();
		Set<AlignmentGroup> selectedMainElements = getSelectedAlignmentGroups(masterJList);
		AlignmentGroup leadMainElement = getLeadSelectedValue(masterJList);

		// Refresh list
		masterJList.getSelectionModel().clearSelection();
		getListView().clear();
		getListView().addAll(newResult.getAlignmentGroups());
		getListView().sort(currentOrdering);

		// Restore previous selection
		for (int i = 0; i < getListView().getListModel().getSize(); i++) {
			AlignmentGroup group = getListView().getListModel().getElementAt(i);
			if (selectedMainElements.contains(group)) {
				getListView().getList().addSelectionInterval(i, i);
			}
			if (leadMainElement != null && group.equals(leadMainElement)) {
				getListView().getList().getSelectionModel().setLeadSelectionIndex(i);
			}
		}
	}

	private ProMTraceList<AlignmentGroup> getListView() {
		return groupedAlignmentsView;
	}

	@SuppressWarnings("unchecked")
	private static <T> T getLeadSelectedValue(@SuppressWarnings("rawtypes") JList list) {
		int leadIndex = list.getLeadSelectionIndex();
		if (leadIndex != -1 && list.getModel().getSize() > leadIndex) {
			return ((T) list.getModel().getElementAt(leadIndex));
		} else {
			return null;
		}
	}

	@SuppressWarnings("unchecked")
	private static <T> Set<T> getSelectedAlignmentGroups(@SuppressWarnings("rawtypes") JList jList) {
		Set<T> selectedValues = new HashSet<>();
		int[] selectedIndices = jList.getSelectedIndices();
		for (int i = 0; i < selectedIndices.length; i++) {
			int index = selectedIndices[i];
			if (index < jList.getModel().getSize()) {
				selectedValues.add(((T) jList.getModel().getElementAt(index)));
			}

		}
		return selectedValues;
	}

	private JLabel createHeading(String label) {
		JLabel title = new JLabel(label);
		title.setAlignmentX(Component.CENTER_ALIGNMENT);
		title.setFont(title.getFont().deriveFont(Font.BOLD));
		title.setForeground(null);
		return title;
	}

}
