package org.processmining.plugins.DataConformance.visualization.grouping;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Set;

import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Event;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Trace;
import org.processmining.plugins.DataConformance.visualization.alignment.AlignmentTrace.DeviationsSetting;
import org.processmining.plugins.DataConformance.visualization.alignment.AlignmentTrace.InvisibleSetting;

import com.google.common.collect.ComparisonChain;

public interface GroupedAlignments<T> {
	
	public enum GroupOrdering implements Comparator<AlignmentGroup> {
		COUNT_ASC("Sort by Count (Ascending)") {
			@Override
			public int compare(final AlignmentGroup a, final AlignmentGroup b) {
				return ComparisonChain.start().compare(a.size(), b.size()).result();
			}
		},
		COUNT_DESC("Sort by Count (Descending)") {
			@Override
			public int compare(final AlignmentGroup a, final AlignmentGroup b) {
				return Collections.reverseOrder(COUNT_ASC).compare(a, b);
			}
		},
		FITNESS_ASC("Sort by Fitness (Ascending)") {
			@Override
			public int compare(final AlignmentGroup a, final AlignmentGroup b) {
				return ComparisonChain.start().compare(a.getAverageFitness(), b.getAverageFitness()).result();
			}
		},
		FITNESS_DESC("Sort by Fitness (Descending)") {
			@Override
			public int compare(final AlignmentGroup a, final AlignmentGroup b) {
				return Collections.reverseOrder(FITNESS_ASC).compare(a, b);
			}
		},
		LENGTH_ASC("Sort by Length (Ascending)") {
			@Override
			public int compare(final AlignmentGroup a, final AlignmentGroup b) {
				return ComparisonChain.start().compare(a.getAverageLength(), b.getAverageLength()).result();
			}
		},
		LENGTH_DESC("Sort by Length (Descending)") {
			@Override
			public int compare(final AlignmentGroup a, final AlignmentGroup b) {
				return Collections.reverseOrder(LENGTH_ASC).compare(a, b);
			}
		};

		private String description;

		private GroupOrdering(String description) {
			this.description = description;
		}

		@Override
		public String toString() {
			return description;
		}
	}
	
	public interface AlignmentGroup {
		
		String getName();
		Trace<? extends Event> getRepresentative(InvisibleSetting invisibleSetting, DeviationsSetting deviationsSetting, boolean colorCodeActivities);
		double getAverageFitness();
		float getAverageLength();
		int size();
		
	}

	void add(T a);
	void addAll(Iterable<T> c);
	int size();
	
	Collection<T> getAlignments();
	Set<AlignmentGroup> getAlignmentGroups();
	Collection<T> listAlignments(AlignmentGroup group);
	int getAlignmentCount(AlignmentGroup group);	

}