package org.processmining.plugins.DataConformance.visualization.grouping;

import java.awt.Color;
import java.util.Collection;
import java.util.Map;
import java.util.Set;

import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMultimap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;

public class GroupedAlignmentsSimpleImpl implements GroupedAlignments<XAlignment> {

	public interface GroupingFunction<F, T> {

		public T apply(F from, GroupedAlignments<XAlignment> groupedAlignments);

	}

	private final Multimap<AlignmentGroup, XAlignment> alignmentMultimap = Multimaps
			.synchronizedListMultimap(ArrayListMultimap.<AlignmentGroup, XAlignment>create());
	private final GroupingFunction<XAlignment, AlignmentGroup> grouping;

	public GroupedAlignmentsSimpleImpl(final Map<String, Color> colorMap) {
		this.grouping = new GroupingFunction<XAlignment, AlignmentGroup>() {

			public AlignmentGroup apply(XAlignment a, GroupedAlignments<XAlignment> groupedAlignments) {
				return new SimpleAlignmentGroup(a, groupedAlignments, colorMap);
			}

		};
	}

	public GroupedAlignmentsSimpleImpl(Iterable<XAlignment> alignments, final Map<String, Color> colorMap) {
		this.grouping = new GroupingFunction<XAlignment, AlignmentGroup>() {

			public AlignmentGroup apply(XAlignment a, GroupedAlignments<XAlignment> groupedAlignments) {
				return new SimpleAlignmentGroup(a, groupedAlignments, colorMap);
			}

		};
		for (XAlignment a : alignments) {
			add(a);
		}
	}

	public GroupedAlignmentsSimpleImpl(GroupingFunction<XAlignment, AlignmentGroup> grouping,
			Iterable<XAlignment> alignments) {
		this.grouping = grouping;
		for (XAlignment a : alignments) {
			add(a);
		}
	}

	public GroupedAlignmentsSimpleImpl(GroupingFunction<XAlignment, AlignmentGroup> grouping) {
		this.grouping = grouping;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.plugins.DataConformance.visualization.grouping.
	 * GroupedResultReplay
	 * #add(org.processmining.plugins.DataConformance.Alignment)
	 */
	@Override
	public void add(XAlignment a) {
		alignmentMultimap.put(grouping.apply(a, this), a);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.plugins.DataConformance.visualization.grouping.
	 * GroupedResultReplay#addAll(java.util.Collection)
	 */
	@Override
	public void addAll(Iterable<XAlignment> c) {
		for (XAlignment a : c) {
			add(a);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.plugins.DataConformance.visualization.grouping.
	 * GroupedResultReplay#getAlignmentGroups()
	 */
	@Override
	public Set<AlignmentGroup> getAlignmentGroups() {
		synchronized (alignmentMultimap) {
			return ImmutableSet.copyOf(alignmentMultimap.keySet());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.plugins.DataConformance.visualization.grouping.
	 * GroupedAlignments
	 * #listAlignments(org.processmining.plugins.DataConformance
	 * .visualization.grouping.GroupedAlignments.AlignmentGroup)
	 */
	@Override
	public Collection<XAlignment> listAlignments(AlignmentGroup group) {
		synchronized (alignmentMultimap) {
			return ImmutableList.copyOf(alignmentMultimap.get(group));
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.plugins.DataConformance.visualization.grouping.
	 * GroupedResultReplay#size()
	 */
	@Override
	public int size() {
		return alignmentMultimap.size();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.plugins.DataConformance.visualization.grouping.
	 * GroupedResultReplay#groupedAlignments()
	 */
	public Multimap<AlignmentGroup, XAlignment> groupedAlignments() {
		synchronized (alignmentMultimap) {
			return ImmutableMultimap.copyOf(alignmentMultimap);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.plugins.DataConformance.visualization.grouping.
	 * GroupedAlignments#getAlignmentCount(org.processmining.plugins.
	 * DataConformance.visualization.grouping.GroupedAlignments.AlignmentGroup)
	 */
	@Override
	public int getAlignmentCount(AlignmentGroup group) {
		synchronized (alignmentMultimap) {
			return alignmentMultimap.keys().count(group);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.plugins.DataConformance.visualization.grouping.
	 * GroupedAlignments#getAlignments()
	 */
	@Override
	public Collection<XAlignment> getAlignments() {
		return alignmentMultimap.values();
	}

}
