package org.processmining.plugins.DataConformance.visualization.grouping;

import java.awt.Color;
import java.text.DecimalFormat;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Map;

import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Event;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Trace;
import org.processmining.plugins.DataConformance.visualization.alignment.AlignmentTrace.DeviationsSetting;
import org.processmining.plugins.DataConformance.visualization.alignment.AlignmentTrace.InvisibleSetting;
import org.processmining.plugins.DataConformance.visualization.alignment.ColorTheme;
import org.processmining.plugins.DataConformance.visualization.grouping.GroupedAlignments.AlignmentGroup;
import org.processmining.xesalignmentextension.XAlignmentExtension.MoveType;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignmentMove;
import org.processmining.xesalignmentextension.XDataAlignmentExtension;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.DataMoveType;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.XDataAlignmentMove;

import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.collect.Iterators;

public class SimpleAlignmentGroup implements AlignmentGroup {

	private final static class CompositeMoveType {

		private final DataMoveType dataMoveType;
		private final MoveType moveType;

		public CompositeMoveType(XDataAlignmentMove move) {
			dataMoveType = move.getDataMoveType();
			moveType = move.getType();
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((dataMoveType == null) ? 0 : dataMoveType.hashCode());
			result = prime * result + ((moveType == null) ? 0 : moveType.hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof CompositeMoveType))
				return false;
			CompositeMoveType other = (CompositeMoveType) obj;
			if (dataMoveType != other.dataMoveType)
				return false;
			if (moveType != other.moveType)
				return false;
			return true;
		}

	}

	private final Map<String, Color> activityColorMap;
	private final GroupedAlignments<XAlignment> groupedAlignments;

	private XAlignment alignment;

	private String[] steps;
	private CompositeMoveType[] types;

	public SimpleAlignmentGroup(XAlignment a, GroupedAlignments<XAlignment> groupedAlignments,
			final Map<String, Color> activityColorMap) {
		this.alignment = a;
		this.groupedAlignments = groupedAlignments;
		this.activityColorMap = activityColorMap;
		this.steps = new String[a.size()];
		this.types = new CompositeMoveType[a.size()];
		int i = 0;
		for (XAlignmentMove move : a) {
			if (move.getType() == MoveType.MODEL) {
				steps[i] = move.getModelMove();
			} else {
				steps[i] = move.getLogMove();
			}
			types[i] = new CompositeMoveType(XDataAlignmentExtension.instance().extendXAlignmentMove(move));
			i++;
		}
	}

	public String getName() {
		return null;
	}

	public int size() {
		return groupedAlignments.getAlignmentCount(this);
	}

	private static final DecimalFormat FLOAT_FORMAT = new DecimalFormat("##0.00%");

	public Trace<? extends Event> getRepresentative(final InvisibleSetting invisibleSetting,
			final DeviationsSetting deviationsSetting, final boolean colorCodeActivities) {
		return new ProMTraceView.Trace<Event>() {

			public String getName() {
				return "Group Size: ".concat(String.valueOf(SimpleAlignmentGroup.this.size()));
			}

			public String getInfo() {
				return "Avg. Fitness: ".concat(FLOAT_FORMAT.format(getAverageFitness()));
			}

			public Iterator<Event> iterator() {
				return Iterators.transform(Iterators.filter(alignment.iterator(), new Predicate<XAlignmentMove>() {

					public boolean apply(XAlignmentMove move) {
						return (invisibleSetting == InvisibleSetting.VISIBLE || move.isObservable());
					}

				}), new Function<XAlignmentMove, Event>() {

					public Event apply(final XAlignmentMove move) {
						final XDataAlignmentMove dataMove = XDataAlignmentExtension.instance().extendXAlignmentMove(
								move);
						return new ProMTraceView.AbstractExtendedEvent() {

							public String getLabel() {
								String modelMove = dataMove.getModelMove();
								if (deviationsSetting == DeviationsSetting.HIGHLIGHTED
										&& dataMove.getType() == MoveType.SYNCHRONOUS) {
									return modelMove != null && modelMove.length() > 0 ? modelMove.substring(0, 1) : modelMove;
								} else {
									return modelMove != null ? modelMove : dataMove
											.getLogMove();
								}
							}

							public Color getWedgeColor() {
								if (colorCodeActivities) {
									if (!move.isObservable()) {
										return Color.LIGHT_GRAY;
									} else {
										String activityId = move.getActivityId();
										if (activityId != null) {
											Color color = activityColorMap.get(activityId);
											if (color != null) {
												return color;	
											} else {
												return Color.LIGHT_GRAY;
											}										
										} else {
											return Color.LIGHT_GRAY;
										}
									}
								} else {
									return ColorTheme.getMoveColor(dataMove);
								}
							}

							public Color getExtendedColor() {
								if (colorCodeActivities) {
									return ColorTheme.getMoveColor(dataMove);
								} else {
									return null;
								}
							}

						};
					}

				});
			}

			public Color getNameColor() {
				return null;
			}

			public Color getInfoColor() {
				return null;
			}

		};
	}

	public double getAverageFitness() {
		return alignment.getFitness();
	}

	public float getAverageLength() {
		return alignment.size();
	}

	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + Arrays.hashCode(steps);
		result = prime * result + Arrays.hashCode(types);
		return result;
	}

	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!(obj instanceof SimpleAlignmentGroup))
			return false;
		SimpleAlignmentGroup other = (SimpleAlignmentGroup) obj;
		if (!Arrays.equals(steps, other.steps))
			return false;
		if (!Arrays.equals(types, other.types))
			return false;
		return true;
	}

}
