/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.cache;

import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.dataflow.PartialDataAlignedTrace;

import it.unimi.dsi.fastutil.objects.Object2ObjectLinkedOpenCustomHashMap;
import it.unimi.dsi.fastutil.shorts.ShortArrays;
import nl.tue.astar.Tail;

/**
 * Caches the resulting {@link PartialDataAlignedTrace} instead of just the data
 * alignment cost. To avoid any overhead you always need to call
 * {@link #setCurrentResult(ControlFlowAlignmentResult)} first and then use the
 * {@link #getResult()}, {@link #contains()} or
 * {@link #putResult(PartialDataAlignedTrace)} methods.
 * <p>
 * PLEASE NOTE: This class is not synchronized!
 * 
 * @author F. Mannhardt
 * 
 */
public class DataViewResultCache extends AbstractDataViewCache {

	private Object2ObjectLinkedOpenCustomHashMap<short[], PartialDataAlignedTrace> resultCache;
	private int maxCacheSize;

	public DataViewResultCache(DataViewBuilder dataViewBuilder, AbstractPDelegate<? extends Tail> delegate,
			boolean activateDataViewCache, int maxCacheSize) {
		super(dataViewBuilder, activateDataViewCache, delegate);
		this.maxCacheSize = maxCacheSize;
		this.resultCache = new Object2ObjectLinkedOpenCustomHashMap<>(Math.max(1, maxCacheSize / 4), ShortArrays.HASH_STRATEGY);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.DataConformance.DataAlignment.PetriNet.Plus
	 * .Cache.DataViewCache#size()
	 */
	@Override
	public int size() {
		return resultCache.size();
	}

	public void putResult(PartialDataAlignedTrace alignmentState) {
		if (activateDataViewCache) {
			if (resultCache.size() > maxCacheSize) {
				resultCache.removeFirst();
			}
			resultCache.put(getExactDataView(), alignmentState);
		}
	}

	public boolean contains() {
		if (!activateDataViewCache) {
			return false;
		}
		return resultCache.containsKey(getExactDataView());
	}

	public PartialDataAlignedTrace getResult() {
		if (!activateDataViewCache) {
			return null;
		}
		return resultCache.getAndMoveToFirst(getExactDataView());
	}

}
