/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.controlflow.adapter;

import nl.tue.astar.Tail;
import nl.tue.astar.impl.AbstractAStarThread.StorageHandler;
import nl.tue.astar.util.FastLookupPriorityQueue;

import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.astar.petrinet.impl.PHead;
import org.processmining.plugins.balancedconformance.cache.DataViewBuilder;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.controlflow.adapter.AlignmentAbstractAdapter.EmptyTraceResult;
import org.processmining.plugins.balancedconformance.controlflow.resultbuilder.AlignmentResultBuilder;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;

/**
 * Interface for building a control flow alignment given a trace with knowledge
 * about the data perspective of the model
 * 
 * @author F. Mannhardt
 * 
 */
public interface AlignmentAdapter extends AutoCloseable {

	public interface PriorityQueueFactory {
		FastLookupPriorityQueue newInstance(DataViewBuilder dataViewBuilder, AbstractPDelegate<? extends Tail> delegate);
	}

	public interface StorageHandlerFactory<T extends Tail> {
		StorageHandler<PHead, T> newStorageHandler(AbstractPDelegate<T> delegate);

		DataStateStore newDataStateStore();
	}

	/**
	 * Returns a {@link ControlFlowAlignmentResultBuilder} that can be used to
	 * retrieve all needed alignments one by one.
	 * 
	 * @param traceIndex
	 * @param traceMapping
	 * @return
	 * @throws ControlFlowAlignmentException
	 */
	AlignmentResultBuilder calcReplayResult(TraceMapping traceMapping, boolean usePartialOrder)
			throws ControlFlowAlignmentException;

	/**
	 * @return for the empty trace (i.e. the cheapest path through the process
	 *         model)
	 * @throws ControlFlowAlignmentException
	 */
	EmptyTraceResult calcEmptyTraceAlignment() throws ControlFlowAlignmentException;

	@Override
	void close();

}
