/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.controlflow.adapter;

import java.util.Map;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.astar.petrinet.impl.PHead;
import org.processmining.plugins.astar.petrinet.impl.PILPTail;
import org.processmining.plugins.balancedconformance.cache.DataViewBuilder;
import org.processmining.plugins.balancedconformance.controlflow.override.DHead;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.balancedconformance.controlflow.override.DataAwareThread;
import org.processmining.plugins.balancedconformance.controlflow.override.queues.duplicates.BFDataDuplicatePriorityQueue;
import org.processmining.plugins.balancedconformance.controlflow.override.queues.duplicates.DFDataDuplicatePriorityQueue;
import org.processmining.plugins.balancedconformance.controlflow.override.queues.duplicates.RandomFirstDuplicatePriorityQueue;
import org.processmining.plugins.balancedconformance.controlflow.resultbuilder.AlignmentResultBuilder;
import org.processmining.plugins.balancedconformance.controlflow.resultbuilder.AlignmentResultBuilderImpl;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapter;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;
import org.processmining.plugins.balancedconformance.observer.DataConformancePlusObserver;

import it.unimi.dsi.fastutil.Hash.Strategy;
import nl.tue.astar.AStarObserver;
import nl.tue.astar.AStarThread.QueueingModel;
import nl.tue.astar.Tail;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.AbstractAStarThread;
import nl.tue.astar.impl.AbstractAStarThread.StorageHandler;
import nl.tue.astar.util.FastLookupPriorityQueue;

/**
 * A* tree based search for the best alignment taking data into account.
 * 
 * @author F. Mannhardt
 * 
 */
public class AlignmentAdapterILPTreeImpl extends AlignmentAdapterILPAbstractImpl {

	public AlignmentAdapterILPTreeImpl(XLog log, DataPetriNet net, LogMapping logMapping, ControlFlowAlignmentConfig c,
			StorageHandlerFactory<PILPTail> storageFactory, boolean isKeepCFSearchSpace, boolean isKeepDFSearchSpace,
			DataConformancePlusObserver dataConformancePlusObserver, DataAlignmentAdapter dataAdapter) {
		super(log, net, logMapping, c, storageFactory, dataConformancePlusObserver, dataAdapter, isKeepCFSearchSpace,
				isKeepDFSearchSpace);
	}

	protected AlignmentResultBuilder createResultBuilder(TraceMapping traceMapping,
			Map<XEvent, XEventClass> mapEv2EventClass, Trace linearTrace, AStarObserver observer) {

		StorageHandler<PHead, PILPTail> localStorageHandler = getStorageHandler();
		DataStateStore localDataStateStore = getDataStore();
		traceMapping.setDataStateStore(localDataStateStore);

		DHead initialHead = new DHead(delegate, initMarking, traceMapping.getTrace());
		AbstractAStarThread<PHead, PILPTail> thread = new DataAwareThread<>(delegate, initialHead, localStorageHandler,
				linearTrace, maxNumOfStates, dataAdapter, logMapping, traceMapping, mapEv2EventClass, mapEvClass2Cost,
				dataViewBuilder, activateCache, useOptimizations, sorting, getQueueFactory(queueingModel), localDataStateStore);

		return new AlignmentResultBuilderImpl((DataAwareThread<PHead, ? extends Tail>) thread, linearTrace,
				traceMapping, canceller, delegate, mapEv2EventClass, mapEvClass2Cost,
				timeLimit, observer);
	}
	
	private PriorityQueueFactory getQueueFactory(QueueingModel queueingModel) {
		switch (queueingModel) {
			case BREADTHFIRST :
				return new PriorityQueueFactory() {
					public FastLookupPriorityQueue newInstance(DataViewBuilder dataViewBuilder,
							AbstractPDelegate<? extends Tail> delegate) {
						return new BFDataDuplicatePriorityQueue(1024 * 2);
					}

				};

			case DEPTHFIRST :
				return new PriorityQueueFactory() {

					public FastLookupPriorityQueue newInstance(DataViewBuilder dataViewBuilder,
							AbstractPDelegate<? extends Tail> delegate) {
						return new DFDataDuplicatePriorityQueue(1024 * 2);
					}

				};

			case RANDOM :
				return new PriorityQueueFactory() {

					public FastLookupPriorityQueue newInstance(DataViewBuilder dataViewBuilder,
							AbstractPDelegate<? extends Tail> delegate) {
						return new RandomFirstDuplicatePriorityQueue(1024 * 2, 0.5);
					}

				};

			default :
				// Fallback
				throw new UnsupportedOperationException("Not supported " + queueingModel);
		}
	}

	protected Strategy<DRecord> getStateGraphStrategy(TraceMapping traceMapping, DataAlignmentAdapter adapter) {
		return new DRecord.DataView.StagedStrategyImpl();
	}

}
