/*
 * Mainly copied from PNetReplayer
 */
package org.processmining.plugins.balancedconformance.controlflow.override;

import gnu.trove.TIntCollection;
import nl.tue.astar.AStarThread;
import nl.tue.astar.Delegate;
import nl.tue.astar.Head;
import nl.tue.astar.Tail;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.AbstractAStarThread;
import nl.tue.astar.impl.State;
import nl.tue.storage.CompressedStore;
import nl.tue.storage.StorageException;

import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataState;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;

public final class DRecordImpl extends AbstractDRecord {

	private long state; //64bit control-flow
	private int dataState = DataStateStore.EMPTY_STATE_INDEX; //32bit data-state

	private long packedCosts; // 21bit control-flow, 21bit data, 21bit estimate	
	private final int packedMoves; // 16bit log move, 16bit model move

	private final int packedSizes; // 16bit numPredecessors, 16bit executed events
	//cached for efficiency, we pay 4 bytes 
	private int hashCode; // 32bit hash code

	private final DRecord predecessor;

	public DRecordImpl(long state, int cost, int dataCost, DRecord predecessor, int logMove, int modelMove,
			int backtraceSize, int executedEvent) {
		this.predecessor = predecessor;
		this.state = state;
		assert PackUtils.isPackableUInt21(cost, dataCost, 0) : "Cannot pack cost " + cost;
		this.packedCosts = PackUtils.packUInt21(cost, dataCost, 0);
		assert PackUtils.isPackableUInt16(backtraceSize, executedEvent) : "Cannot pack backtrace " + backtraceSize
				+ " or executedEvent " + backtraceSize;
		this.packedSizes = PackUtils.packUInt16(backtraceSize, executedEvent);
		int packedLog = logMove == AStarThread.NOMOVE ? PackUtils.MAX_UINT16_PACKABLE : logMove;
		int packedModel = modelMove == AStarThread.NOMOVE ? PackUtils.MAX_UINT16_PACKABLE : modelMove;
		assert PackUtils.isPackableUInt16(packedLog, packedModel) : "Cannot pack log move " + packedLog
				+ " or model move " + packedModel;
		this.packedMoves = PackUtils.packUInt16(packedLog, packedModel);
	}

	public DRecordImpl(int cost, int dataCost, DRecord predecessor) {
		this.predecessor = predecessor;
		this.packedCosts = PackUtils.packUInt21(cost, dataCost, 0);
		this.packedMoves = PackUtils.packUInt16(PackUtils.MAX_UINT16_PACKABLE, PackUtils.MAX_UINT16_PACKABLE);
		this.packedSizes = PackUtils.packUInt16(PackUtils.MAX_UINT16_PACKABLE, 0);
	}

	@Override
	public <H extends Head, T extends Tail> State<H, T> getState(CompressedStore<State<H, T>> storage)
			throws StorageException {
		return storage.getObject(getState());
	}

	@Override
	public int getCostSoFar() {
		if (getDataCost() == Integer.MAX_VALUE) {
			return Integer.MAX_VALUE;
		}
		return getCost() + getDataCost();
	}

	@Override
	public DRecord getPredecessor() {
		return predecessor;
	}

	@Override
	public double getTotalCost() {
		int dataCost = getDataCost();
		if (dataCost == Integer.MAX_VALUE) {
			return Integer.MAX_VALUE;
		}
		return getCost() + dataCost + getEstimateAsInt();
	}

	@Override
	public int getTotalIntegerCost() {
		int dataCost = getDataCost();
		if (dataCost == Integer.MAX_VALUE) {
			return Integer.MAX_VALUE;
		}
		return getCost() + dataCost + getEstimateAsInt();
	}

	@Override
	public long getState() {
		return state;
	}

	@Override
	public void setState(long state) {
		if (state < 0) {
			throw new RuntimeException("Can only store non-negative states!");
		}
		this.state = state;
		this.hashCode = 0; //invalidate hash since state changed
	}

	@Override
	public DRecord getNextRecord(Delegate<? extends Head, ? extends Tail> d, Trace trace, Head nextHead, long state,
			int modelMove, int movedEvent, int activity) {
		AbstractPDelegate<? extends Tail> delegate = (AbstractPDelegate<?>) d;
		assert !(modelMove != AStarThread.NOMOVE && movedEvent != AStarThread.NOMOVE)
				|| delegate.getActivitiesFor((short) modelMove).contains((short) activity);
		int c = delegate.getCostFor(modelMove, activity);

		int newExecuted = getExecutedEvents();
		if (movedEvent != AStarThread.NOMOVE) {
			newExecuted++;
		}

		return new DRecordImpl(state, getCost() + c, getDataCost(), this, movedEvent, modelMove,
				getBacktraceSize() + 1, newExecuted);
	}

	@Override
	public boolean isExactEstimate() {
		return true;
	}

	@Override
	public double getEstimatedRemainingCost() {
		return getEstimateAsInt();
	}

	private int getEstimateAsInt() {
		int estimateAsInt = PackUtils.unpackUInt21Left(packedCosts);
		if (estimateAsInt == PackUtils.MAX_UINT21_PACKABLE) {
			return -1; // ESTIMATEIRRELEVANT was stored
		}
		return estimateAsInt;
	}

	@Override
	public void setEstimatedRemainingCost(double cost, boolean isExactEstimate) {
		// Only supporting integer estimates
		this.packedCosts = PackUtils.changeUInt21Left(packedCosts, (int) cost);
	}

	@Override
	public int getModelMove() {
		int move = PackUtils.unpackUInt16Left(packedMoves);
		return move == PackUtils.MAX_UINT16_PACKABLE ? AStarThread.NOMOVE : move;
	}

	@Override
	public int getMovedEvent() {
		int move = PackUtils.unpackUInt16Right(packedMoves);
		return move == PackUtils.MAX_UINT16_PACKABLE ? AStarThread.NOMOVE : move;
	}

	@Override
	public TIntCollection getNextEvents(Delegate<? extends Head, ? extends Tail> delegate, Trace trace) {
		boolean[] executed = new boolean[trace.getSize()];
		fillExecuted(executed);
		return trace.getNextEvents(executed);
		//		throw new UnsupportedOperationException("DRecord only supports linear traces!");
	}

	private void fillExecuted(boolean[] executed) {
		DRecordImpl rec = this;
		do {
			if (rec.getMovedEvent() != AStarThread.NOMOVE) {
				executed[rec.getMovedEvent()] = true;
			}
			rec = (DRecordImpl) rec.getPredecessor();
		} while (rec != null);
	}

	@Override
	public int getNextEvent(Delegate<? extends Head, ? extends Tail> delegate, Trace trace) {
		return getExecutedEvents();
	}

	@Override
	public int getBacktraceSize() {
		int backTrace = PackUtils.unpackUInt16Right(packedSizes);
		// -1 for root, 0 for first layer, 1 for second layer, ... 
		return backTrace == PackUtils.MAX_UINT16_PACKABLE ? -1 : backTrace;
	}

	public int getExecutedEvents() {
		return PackUtils.unpackUInt16Left(packedSizes);
	}

	@Override
	public int getCost() {
		return PackUtils.unpackUInt21Right(packedCosts);
	}

	@Override
	public int getDataCost() {
		return PackUtils.unpackUInt21Middle(packedCosts);
	}

	public void setDataCost(int dataCost) {
		this.packedCosts = PackUtils.changeUInt21Middle(packedCosts, dataCost);
	}

	@Override
	public DataState getDataState(DataStateStore store) {
		return store.get(getDataState());
	}

	public void setDataState(int dataState) {
		//assert dataState >= 0 : "Can only store non-negative states!";
		this.dataState = dataState;
	}

	@Override
	public int getDataState() {
		return dataState;
	}

	public boolean equals(Object obj) {
		throw new UnsupportedOperationException("Use DRecord.DataView.StrategyImpl instead!");
	}

	public int hashCode() {
		throw new UnsupportedOperationException("Use DRecord.DataView.StrategyImpl instead!");
	}

	public void setHashCode(int hashCode) {
		this.hashCode = hashCode;
	}

	public int getHashCode() {
		return hashCode;
	}

	public String toString() {
		return "[s:" + getState() + " m:" + ((getModelMove() != AbstractAStarThread.NOMOVE) ? getModelMove() : -1)
				+ " e:" + ((getMovedEvent() != AbstractAStarThread.NOMOVE) ? getMovedEvent() : -1) + " c:" + getCost()
				+ " e:" + getEstimatedRemainingCost() + "]";
	}

}