/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.dataflow.datastate;

/**
 * Stores {@link DataState} under an index.
 * 
 * @author F. Mannhardt
 *
 */
public interface DataStateStore {
	
	public static final int START_INDEX = Integer.MIN_VALUE+2;
	public static final int EMPTY_STATE_INDEX = Integer.MIN_VALUE+1;
	public static final Integer NO_ENTRY = Integer.MIN_VALUE;
	
	/**
	 * Empty states that is immutable, i.e. all mutating methods throw
	 * {@link UnsupportedOperationException}.
	 */
	public static final DataState EMPTY_IMMUTABLE_STATE = new EmptyDataState();

	/**
	 * Returns the {@link DataState} that is stored under the specified index. 
	 * 
	 * @param index
	 * @return {@link DataState} or NULL, if there is none
	 */
	DataState get(int index);

	/**
	 * Stores the {@link DataState} and returns the index that it is stored under. 
	 * 
	 * @param state
	 * @return
	 */
	int putIfAbsent(DataState state);

	/**
	 * @return the number of {@link DataState} in this store
	 */
	int size();

}
