package org.processmining.plugins.balancedconformance.dataflow.datastate;

import it.unimi.dsi.fastutil.ints.Int2ObjectMap;
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;

public class DataStateStoreNoLockImpl implements DataStateStore {

	private int stateCounter = START_INDEX;

	private final Object2IntMap<DataState> dataStateStore;
	private final Int2ObjectMap<DataState> reverseDataStateStore;

	public DataStateStoreNoLockImpl(int initialCapacity) {
		super();
		dataStateStore = new Object2IntOpenHashMap<>(initialCapacity, 0.5f);
		dataStateStore.defaultReturnValue(NO_ENTRY);
		reverseDataStateStore = new Int2ObjectOpenHashMap<>(initialCapacity, 0.5f);
	}

	public DataState get(int index) {
		if (index == EMPTY_STATE_INDEX) {
			return EMPTY_IMMUTABLE_STATE;
		}
		return reverseDataStateStore.get(index);
	}

	public int putIfAbsent(DataState state) {
		if (state == DataStateStore.EMPTY_IMMUTABLE_STATE) {
			return EMPTY_STATE_INDEX;
		}
		state = state.makeStateImmutable();
		int idx = dataStateStore.getInt(state);
		if (idx != NO_ENTRY) {
			return idx;
		} else {
			int stateId = stateCounter++;
			dataStateStore.put(state, stateId);
			reverseDataStateStore.put(stateId, state);
			return stateId;
		}
	}

	public int size() {
		return dataStateStore.size();
	}

}
