package org.processmining.plugins.balancedconformance.dataflow.util;

import java.util.Map;

import org.processmining.datapetrinets.exception.EvaluatorException;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataState;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizer;

/**
 * Evaluates an {@link GuardExpression} looking up values for variables in the
 * supplied {@link DataState}.
 * 
 * @author F. Mannhardt
 * 
 */
public final class ExpressionEvaluator {
	
	public static boolean failSafeIsFalse(GuardExpression expression, DataState variableValues, @SuppressWarnings("rawtypes") Map<String, Class> variableTypes,
			StringDiscretizer discretizer) {
		boolean allVariablesNotNull = checkAllVariableNotNull(expression, variableValues);
		return allVariablesNotNull && isFalse(expression, variableValues, variableTypes, discretizer);
	}
	
	public static boolean failSafeIsTrue(GuardExpression expression, DataState variableValues, @SuppressWarnings("rawtypes") Map<String, Class> variableTypes,
			StringDiscretizer discretizer) {
		boolean allVariablesNotNull = checkAllVariableNotNull(expression, variableValues);
		return allVariablesNotNull && isTrue(expression, variableValues, variableTypes, discretizer);
	}

	public static boolean checkAllVariableNotNull(GuardExpression expression, DataState variableValues) {
		boolean allVariablesAvailable = true;
		for (String varName: expression.getNormalVariables()) {
			allVariablesAvailable &= variableValues.contains(varName);
		}
		for (String varName: expression.getPrimeVariables()) {
			allVariablesAvailable &= variableValues.contains(DataStateHelper.addPrime(varName));
		}
		return allVariablesAvailable;
	}

	public static boolean isTrue(GuardExpression expression, DataState variableValues, @SuppressWarnings("rawtypes") Map<String, Class> variableTypes,
			StringDiscretizer discretizer) throws EvaluatorException {
		
		Object o = expression.evaluate(new DataStateVariableProvider(variableValues, discretizer, variableTypes));

		if (o instanceof Boolean) {
			return ((Boolean) o).booleanValue();
		} else {
			throw new IllegalArgumentException("Expression " + expression.toCanonicalString()
					+ " evaluates to a non-boolean result " + o);
		}
		
	}

	public static boolean isFalse(GuardExpression expression, DataState variableValues, @SuppressWarnings("rawtypes") Map<String, Class> variableTypes,
			StringDiscretizer discretizer) throws EvaluatorException {
		return (!isTrue(expression, variableValues, variableTypes, discretizer));
	}

}
