/*
 * Copyright (c) 2014 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.export;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.log.utils.XUtils;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.ResultReplay;
import org.processmining.plugins.balancedconformance.result.AlignmentCollection;
import org.processmining.plugins.balancedconformance.result.BalancedReplayResult;
import org.processmining.xesalignmentextension.XAlignmentExtension;
import org.processmining.xesalignmentextension.XDataAlignmentExtension;

/**
 * Export Alignments to a 'repaired' XLog with the XAlignmentExtension
 * 
 * @author F. Mannhardt
 * 
 */
public class ExtendLogWithAlignments {

	@Plugin(name = "Convert Data Alignment to Event Log with XAlignmentExtension", parameterLabels = { "AlignmentCollection" }, level = PluginLevel.NightlyBuild, returnLabels = { "Log" }, returnTypes = { XLog.class }, userAccessible = true,
			help = "Convert Data Alignment to Event Log with XAlignmentExtension for exporting and storing alignments.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl", pack = "DataAwareReplayer")
	public XLog extendLogWithAlignments(PluginContext context, AlignmentCollection alignments) {

		context.getProgress().setCaption("Creating new Log with XAlignmentExtension ...");

		XLog extendedLog = doExtendLogWithAlignments(context, alignments);

		context.getProgress().setCaption("Finished!");

		return extendedLog;

	}

	public static XLog doExtendLogWithAlignments(PluginContext context, AlignmentCollection alignments) {
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog extendedLog = factory.createLog();

		XAlignmentExtension alignmentExtension = XAlignmentExtension.instance();
		XDataAlignmentExtension dataAlignmentExtension = XDataAlignmentExtension.instance();
		XConceptExtension conceptExtension = XConceptExtension.instance();
		extendedLog.getExtensions().add(conceptExtension);
		extendedLog.getExtensions().add(alignmentExtension);
		extendedLog.getExtensions().add(dataAlignmentExtension);

		XAlignmentConverter converter = new XAlignmentConverter().setFactory(factory);
		
		context.getProgress().setMaximum(alignments.getAlignments().size());
		
		if (alignments instanceof BalancedReplayResult) {
			
			converter.setClassifier(((ResultReplay) alignments).getClassifier());
			converter.setVariableMapping(((ResultReplay) alignments).getVariableMapping());
			
			alignmentExtension.assignFitness(extendedLog, ((ResultReplay)alignments).meanFitness);
			
			for (XTrace trace : ((ResultReplay) alignments).getAlignedLog()) {

				String traceName = XUtils.getConceptName(trace);
				Alignment a = ((ResultReplay) alignments).getAlignmentByTraceName(traceName);
				
				if (a != null) {
					XTrace newTrace = converter.copyAsXTrace(a, trace);
					extendedLog.add(newTrace);					
				} else {
					context.log("WARN: Trace "+traceName + " unable to find alignment!" );
				}
				
				context.getProgress().inc();
			}
			
		} else {
			for (Alignment a: alignments) {
				extendedLog.add(converter.copyAsXTrace(a));
				context.getProgress().inc();
			}
		}
		return extendedLog;
	}
}