/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.result;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult.Step;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.petrinet.replayresult.StepTypes;

/**
 * Calculates the maximum cost for a given {@link XTrace} to determine the
 * fitness.
 * 
 * @author F. Mannhardt
 *
 */
public class MaxAlignmentCostHelper {

	private final ControlFlowAlignmentResult emptyTraceAlignment;
	private final XEventClasses eventClasses;
	private final Map<XEventClass, Integer> mapEvClass2Cost;
	private final Map<Transition, Integer> mapTrans2Cost;
	private final Map<XEventClass, Transition> mapEvClass2Trans;
	private final VariableMatchCosts variableCost;
	private float maxMoveInModelCost;

	public MaxAlignmentCostHelper(ControlFlowAlignmentResult emptyTraceAlignment, TransEvClassMapping transEvMapping,
			XEventClasses eventClasses, Map<XEventClass, Integer> mapEvClass2Cost,
			Map<Transition, Integer> mapTrans2Cost, VariableMatchCosts variableCost) {
		super();
		this.emptyTraceAlignment = emptyTraceAlignment;
		this.eventClasses = eventClasses;
		this.mapEvClass2Cost = mapEvClass2Cost;
		this.mapTrans2Cost = mapTrans2Cost;
		this.variableCost = variableCost;
		this.mapEvClass2Trans = new HashMap<>();
		for (Entry<Transition, XEventClass> entry : transEvMapping.entrySet()) {
			mapEvClass2Trans.put(entry.getValue(), entry.getKey());
		}
		this.maxMoveInModelCost = getMaxMoveInModelCost();
	}

	public float getMaxCost(XTrace trace) {
		float moveInLogMaxCost = 0.0f;
		for (XEvent ev : trace) {
			XEventClass classOf = eventClasses.getClassOf(ev);
			moveInLogMaxCost += mapEvClass2Cost.get(classOf);
		}

		return maxMoveInModelCost + moveInLogMaxCost;
	}

	private float getMaxMoveInModelCost() {
		float moveInModelMaxCost = 0.0f;
		Iterator<Step> stepIter = emptyTraceAlignment.getSteps().iterator();
		while (stepIter.hasNext()) {
			Step step = stepIter.next();
			if (step.getType() == StepTypes.MREAL) {
				//TODO consider virtual variables!!!
				Transition node = step.getModelNode();
				moveInModelMaxCost += (calcMaxDataCost(node) + mapTrans2Cost.get(step.getModelNode()));
			}

		}
		return moveInModelMaxCost;
	}

	private float calcMaxDataCost(Transition node) {
		float maxDataCost = 0;
		if (node instanceof PNWDTransition) {
			for (DataElement el : ((PNWDTransition) node).getWriteOperations()) {
				maxDataCost += Math.max(variableCost.notWritingCost(node.getLabel(), el.getVarName()),
						variableCost.costFaultyValue(node.getLabel(), el.getVarName()));
			}
		}
		return maxDataCost;
	}

}